#!/bin/env python3
# -*- coding: utf-8 -*-

"""
    Linear Advance 1.5 pattern generator, version 2022-08-17
    Copyright (C) 2022  A. Steiner

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

"""
                                    SUMMARY

    This script will generate a calibration pattern (like the one from marlin)
    to help to calibrate the linear advance 1.5.
    From what I have understood, the K factor will give the length of filament
    that will be retracted/unretracted during the deceleration/acceleration
    phases of the hotend: just multiply it by the speed of the extruder (not
    the hotend!) to get the length. This speed has to be computed.
    Since this length is proportional to the speed of the extruder, it is
    longer for fast speed and shorter for slow speed. Be careful with the
    retraction settings! (See below)
    It depends also on the layer height: a small layer height should requires
    less extrusion (for the same speed), and so a lower K factor. But for large
    nozzles it can be the opposite since they are more leaky. The calibration
    should therefore be done for all the layer heights used.
    Keep in mind that the filament is retracted during the deceleration phase:
    if a retraction occurs after that, it will be added, which could lead to a
    big amount of total retraction for the fast speed movements! The script
    will adapt the retraction length accordingly to stay under the limit set
    and avoid clogs. It will also print all the retractions lengths to have
    them in check.
    The K factor depends also on all the parameters that affect the extrusion,
    like the temperature used, the filament brand and type, etc...
    In summary:
        - set the MAX_SPEED as your maximum speed used;
        - check the retraction length for the best K factor;
        - if this length is not enough to avoid strings, use a lower speed or
          layer height. According to the doc, only a small amount of retraction
          is required if the K factor is well calibrated.
"""

# Get the global settings from prusasslicer
# The speeds should be small perimeters / perimeters for instance

# Add a thumbnail? (require base64, cv2, numpy)
# They can be installed with the command:
#       python3 -m pip install b64 opencv-python numpy
THUMBNAIL = True

# =============================== Part to edit ==============================
# Print parameters
NOZZLE_DIAMETER = 0.6
LINE_WIDTH = 0.65
LAYER_HEIGHT = 0.20
EXTRUSION_FLOW = 1

# Parameters of the K factor
K_MIN = 0
K_MAX = 0.5
K_INC = 0.1

# Temperatures
BED_TEMP = 50
NOZ_TEMP = 195

# Retraction
RETRACT_LENGTH = 2.4
RETRACT_SPEED = 70
UNRETRACT_SPEED = 40

# Speeds
MIN_SPEED = 15
MAX_SPEED = 70
TRAVEL_SPEED = 150

# Output
GCODE = "LA15_calibration_pattern_for_MINI.gcode"
# ============================ End of part to edit ===========================

import sys
orig_stdout = sys.stdout

# Parameters of the lines
SMALL_LENGTH = 20
BIG_LENGTH = 50
X0 = 50
X1 = X0 + SMALL_LENGTH
X2 = X1 + BIG_LENGTH
X3 = X2 + SMALL_LENGTH
Y0 = 20
Y_MAX = 150  # safeguard
Y_STEP = 5

# Flow
CROSS_SECTION = (1.75/2)**2 * 3.14159
EXTRUDED_SECTION = (LAYER_HEIGHT/2)**2 * 3.14159 \
                   + LAYER_HEIGHT * (LINE_WIDTH - LAYER_HEIGHT)


# Compute the extrusion length
def extrusion_length(distance):
    return EXTRUDED_SECTION / CROSS_SECTION * (distance * EXTRUSION_FLOW)


# Linear advance retract the filament, so we compute the length
# to retract to stay under RETRACT_LENGTH
def la_retraction_length(speed, k_factor):
    return (EXTRUDED_SECTION / CROSS_SECTION * EXTRUSION_FLOW * k_factor * speed)


def retraction_length(speed, k_factor):
    length = RETRACT_LENGTH - la_retraction_length(speed, k_factor)
    if length <= 0:
        sys.stdout = orig_stdout
        print("")
        print("ERROR:")
        print("       the K factor %.2F is too high: the retraction will be above %smm (%.5f)" % (k_factor, RETRACT_LENGTH, la_retraction_length(speed, k_factor)))
        exit(-1)
    return length


# The extruded length used
SMALL_EXTRUSION_LENGTH = extrusion_length(SMALL_LENGTH)
BIG_EXTRUSION_LENGTH = extrusion_length(BIG_LENGTH)

# Init
K = K_MIN
Y = Y0
i = 0

# Filament length (in mm)
FILAMENT_LENGTH = 0

###############################################################################
#                                 Write of gcode
# ----------------------------------------------------------------------------

# Add progress values
percent = 0
progress = 100 / ((K_MAX - K_MIN) / K_INC + 8)


def print_progress():
    global percent
    print("M73 P%s R0" % (round(percent)))
    percent += progress


with open(GCODE, 'w') as f:
    sys.stdout = f

    # Show parameters
    print("; Parameters used:")
    print(";  - nozzle: %smm" % (NOZZLE_DIAMETER))
    print(";  - line width: %.2f" % (LINE_WIDTH))
    print(";  - layer height: %.2f" % (LAYER_HEIGHT))
    print(";  - nozzle temperature: %s°C" % (NOZ_TEMP))
    print(";  - first K: %.3f" % (K_MIN))
    print(";  - last K: %.3f" % (K_MAX))
    print(";  - low speed: %smm/s" % (MIN_SPEED))
    print(";  - high speed: %smm/s" % (MAX_SPEED))
    print("")

    # Thumbnail
    if THUMBNAIL:
        import base64
        import cv2
        import numpy

        line1 = "%s@%s" % (NOZZLE_DIAMETER, LAYER_HEIGHT)
        line2 = "K %s-%s" % (K_MIN, K_MAX)
        font = cv2.FONT_HERSHEY_SIMPLEX
        img = numpy.zeros((124, 220, 3), numpy.uint8)
        img[:, :] = (86, 86, 86)
        color = (70, 116, 219)
        img_txt = cv2.putText(img, line1, (0, 40), font, 1, color, 2, cv2.LINE_AA)
        img_txt = cv2.putText(img, line2, (0, 80), font, 1, color, 2, cv2.LINE_AA)
        png = cv2.imencode('.png', img_txt)
        thumbnail = base64.b64encode(png[1]).decode('utf-8')
        # uncomment if you want to see the image
        # cv2.imwrite("/tmp/test.png", img_txt)

        print(";")
        print("; thumbnail begin 220x124 %s" % (len(thumbnail)))
        j = 0
        while j < len(thumbnail):
            print("; %s" % (thumbnail[j:j+78]))
            j += 78
        print("; thumbnail end")
        print(";")
        print("")

    # Before print
    print(";=== Before print ====================================================")
    print_progress()
    print("M201 X2500 Y2500 Z400 E5000 ; sets maximum accelerations, mm/sec^2")
    print("M203 X180 Y180 Z12 E80 ; sets maximum feedrates, mm / sec")
    print("M204 P1250 R1250 T2500 ; sets acceleration (P, T) and retract acceleration (R), mm/sec^2")
    print("M205 X8.00 Y8.00 Z2.00 E10.00 ; sets the jerk limits, mm/sec")
    print("M205 S0 T0 ; sets the minimum extruding and travel feed rate, mm/sec")
    print("M107")
    print(";TYPE:Custom")
    print("M862.3 P \"MINI\" ; printer model check")
    print("G90 ; use absolute coordinates")
    print("M83 ; extruder relative mode")
    print("M104 S170 ; set extruder temp for bed leveling")
    print("M140 S%s ; set bed temp" % (BED_TEMP))
    print("M109 R170 ; wait for extruder leveling temp")
    print("M190 S%s ; wait for bed temp" % (BED_TEMP))
    print("M204 T1250 ; set travel acceleration")
    print("G28 ; home all without mesh bed level")
    print("G29 ; mesh bed leveling ")
    print("M204 T2500 ; restore travel acceleration")
    print("M104 S%s ; set extruder temp" % (NOZ_TEMP))
    print("G92 E0")
    print("G1 Y-2 X179 F2400")
    print_progress()
    print("G1 Z3 F720")
    print("M109 S%s ; wait for extruder temp" % (NOZ_TEMP))
    print("")
    print("; intro line")
    print_progress()
    print("G1 X170 F1000")
    print_progress()
    print("G1 Z0.2 F720")
    print_progress()
    print("G1 X110 E8 F900")
    print("G1 X40 E10 F700")
    print("G92 E0")
    print("M221 S95 ; set flow")
    print("")
    print("G1 E-%.2f F%s; retract" % (RETRACT_LENGTH, RETRACT_SPEED * 60))
    print("G1 Z%.2f F720" % (LAYER_HEIGHT))
    print("")
    print("G21 ; set units to millimeters")
    print("G90 ; use absolute coordinates")
    print("M83 ; use relative distances for extrusion")
    print("M107")
    print("G92 E0.0")
    print("M204 P800")
    print("")
    print("")
    FILAMENT_LENGTH += 18

    # Loop
    print(";==== K Lines ========================================================")
    R_LENGTH = RETRACT_LENGTH
    while K <= K_MAX:
        i = i+1
        # the retraction length to use
        R_LENGTH = retraction_length(MIN_SPEED, K)
        # check if the max speed respect the retraction length with this K factor
        dump = retraction_length(MAX_SPEED, K)
        print("; Line %d : K=%.2f" % (i, K))
        print("; Min speed LA retraction: %.5f" % (la_retraction_length(MIN_SPEED, K)))
        print("; Max speed LA retraction: %.5f" % (la_retraction_length(MAX_SPEED, K)))
        print(";TYPE:External perimeter")
        print_progress()
        print("G1 Z%.2f" % (LAYER_HEIGHT + 0.2))
        print("G1 X%.2f Y%.2f F%s" % (X0, Y, TRAVEL_SPEED * 60))
        print("M900 K%.2f ; set K-factor" % (K))
        print("G1 Z%.2f" % (LAYER_HEIGHT))
        print("G1 E%.5f F%s; un-retract" % (R_LENGTH, UNRETRACT_SPEED * 60))
        print("G1 X%.2f Y%.2f E%.5f F%s ; print line" % (X1, Y, SMALL_EXTRUSION_LENGTH, MIN_SPEED * 60))
        print("G1 X%.2f Y%.2f E%.5f F%s ; print line" % (X2, Y, BIG_EXTRUSION_LENGTH, MAX_SPEED * 60))
        print("G1 X%.2f Y%.2f E%.5f F%s ; print line" % (X3, Y, SMALL_EXTRUSION_LENGTH, MIN_SPEED * 60))
        print("G1 E-%.5f F%s; retract" % (R_LENGTH, RETRACT_SPEED * 60))
        print("")
        FILAMENT_LENGTH += 2*SMALL_EXTRUSION_LENGTH + BIG_EXTRUSION_LENGTH

        # Set the next step
        K = round(K + K_INC, 3)
        Y = round(Y + Y_STEP, 3)
        if (Y >= Y_MAX):
            break

    # Join the lines
    Y = round(Y - Y_STEP, 3)
    JOIN_EXTRUSION_LENGTH = extrusion_length(Y - Y0)
    print(";=== Join lines ======================================================")
    print("; Joining the lines")
    print("M900 K0 ; reset K factor: we print at slow speed now")
    print(";TYPE:External perimeter")
    print_progress()
    print("G1 X%.2f Y%.2f F%s" % (X3, Y, TRAVEL_SPEED * 60))
    print("G1 E%.2f F%s; un-retract" % (RETRACT_LENGTH, UNRETRACT_SPEED * 60))
    print("G1 X%.2f Y%.2f E%.5f F%s" % (X3, Y0, JOIN_EXTRUSION_LENGTH, MIN_SPEED * 60))
    print("G1 E-%.2f F%s; retract" % (RETRACT_LENGTH, RETRACT_SPEED * 60))
    print("G1 Z%.2f" % (LAYER_HEIGHT + 0.2))
    print("G1 X%.2f Y%.2f F%s" % (X0+LINE_WIDTH/2, Y0, TRAVEL_SPEED * 60))
    print(";TYPE:External perimeter")
    print_progress()
    print("G1 Z%.2f" % (LAYER_HEIGHT))
    print("G1 E%.2f F%s; un-retract" % (RETRACT_LENGTH, UNRETRACT_SPEED * 60))
    print("G1 X%.2f Y%.2f E%.5f F%s" % (X0+LINE_WIDTH/2, Y, JOIN_EXTRUSION_LENGTH, MIN_SPEED * 60))
    print("G1 E-%.2f F%s; retract" % (RETRACT_LENGTH, RETRACT_SPEED * 60))
    print("")
    FILAMENT_LENGTH += 2*JOIN_EXTRUSION_LENGTH

    # After print
    print(";=== After print =====================================================")
    print(";TYPE:Custom")
    print_progress()
    print("G1 Z2 F720 ; Move print head up")
    print("G1 X178 Y178 F4200 ; park print head")
    print("G1 Z40 F720 ; Move print head further up")
    print("G4 ; wait")
    print("M104 S0 ; turn off temperature")
    print("M140 S0 ; turn off heatbed")
    print("M107 ; turn off fan")
    print("M221 S100 ; reset flow")
    print("M900 K0 ; reset LA")
    print("M84 ; disable motors")
    print("M73 P100 R0")
    print("")
    print("; filament used [mm] = %.2f" % (FILAMENT_LENGTH))
    sys.stdout = orig_stdout

print("The gcode has been written info %s" % (GCODE))
